// © Cosmos Software | Ali Yigit Bireroglu All material used in the making of this code, project, program, application,
// software et cetera (the "Intellectual Property") belongs completely and solely to Ali Yigit Bireroglu. This includes but
// is not limited to the source code, the multimedia and other asset files. If you were granted this Intellectual Property
// for personal use, you are obligated to include this copyright text at all times.
// Copyright © 2019 Ali Yigit Bireroglu (https://www.cosmossoftare.coffee) All rights reserved.

//@formatter:off
import 'dart:ui';
import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter/services.dart';
import 'gesture_detector.dart' as MyGestureDetector;

import 'Export.dart';

///The widget that is responsible of ALL Peek & Pop related logic. It works together with [PeekAndPopChildState] and [PeekAndPopDetector].
class PeekAndPopController extends StatefulWidget {
  ///The widget that is to be displayed on your regular UI.
  final Widget uiChild;
  ///The view to be displayed during the Peek & Pop process.
  final PeekAndPopBuilder peekAndPopBuilder;
  ///Set this to true if your [peekAndPopBuilder] doesn't change during the Peek & Pop process.
  final bool useCache;
  ///The color to be displayed over the [Backdrop] widget. The alpha of this color will be controlled internally and it will not exceed 126 (0.5).
  final Color backdropColor;
  ///An optional second view to be displayed over the [Backdrop] widget during the Peek & Pop process.
  final Widget overlayBuiler;
  ///Set this to true if your [peekAndPopBuilder] uses a [Hero] widget.
  final bool isHero;
  
  ///The callback for when the Peek & Pop successfully completes. This callback is invoked after the entire process.
  final PeekAndPopCallback onPeekAndPopComplete;
  ///The callback for when the view is initially pushed to the Navigator.
  final PeekAndPopCallback onPushPeekAndPop;
  ///The callback for when the state of the view is updated.
  final PeekAndPopCallback onUpdatePeekAndPop;
  ///The callback for when the Peek & Pop is cancelled.
  final PeekAndPopCallback onCancelPeekAndPop;
  ///The callback for when the view is being popped after peeking.
  final PeekAndPopCallback onFinishPeekAndPop;
  ///The callback for when the view is being closed after popping.
  final PeekAndPopCallback onClosePeekAndPop;

  ///The pressure treshold at which the view first becomes visible. The [Backdrop] widget does not depend on this.
  final double treshold;
  ///The min pressure for initiating the Peek & Pop process. Flutter's normal [GestureDetector] has relatively wrong values for a Peek & Pop process
  ///so a new [MyGestureDetector.GestureDetector] is created for altering these values. The new [MyGestureDetector.GestureDetector] is otherwise
  ///identical to Flutter's normal [GestureDetector].
  final double startPressure;
  ///The max pressure for finishing the Peek & Pop process. Flutter's normal [GestureDetector] has relatively wrong values for a Peek & Pop process
  ///so a new [MyGestureDetector.GestureDetector] is created for altering these values. The new [MyGestureDetector.GestureDetector] is otherwise
  ///identical to Flutter's normal [GestureDetector].
  final double peakPressure;
  ///The scale at which you want your view to be displayed when it first becomes visible.
  final double peekScale;
  ///The coefficient for controlling how much your view can rescale according to the pressure magnitude. A low value such as 0.05 is 
  ///recommended based on observations made by Swiss scientists.
  final double peekCoefficient;
  ///See [PeekAndPopMoveController].
  final PeekAndPopMoveController moveController;
  
  ///The transition to be used when:
  ///   a) The view is opened directly 
  ///   b) The view is closed. A default [SlideTransition] is provided.
  final Function popTransition;
  //TODO Change the name of this variable as it can cause confusion.

  const PeekAndPopController(
          this.uiChild,
          this.peekAndPopBuilder,
          this.useCache,
          {this.backdropColor: Colors.black,
          this.overlayBuiler,
          this.isHero:false,
          this.onPeekAndPopComplete,
          this.onPushPeekAndPop,
          this.onUpdatePeekAndPop,
          this.onCancelPeekAndPop,
          this.onFinishPeekAndPop,
          this.onClosePeekAndPop,
          this.treshold: 0.5,
          this.startPressure: 0.2,
          this.peakPressure: 0.9,
          this.peekScale: 0.5,
          this.peekCoefficient: 0.05,
          this.moveController:null,  
          this.popTransition});

  @override
  PeekAndPopControllerState createState() {
    return PeekAndPopControllerState(
            uiChild,
            peekAndPopBuilder,
            useCache,
            backdropColor,
            overlayBuiler,
            isHero,
            onPeekAndPopComplete,
            onPushPeekAndPop,
            onUpdatePeekAndPop,
            onCancelPeekAndPop,
            onFinishPeekAndPop,
            onClosePeekAndPop,
            treshold,
            startPressure,
            peakPressure,
            peekScale,
            peekCoefficient,
            moveController,
            popTransition);
  }
}

class PeekAndPopControllerState extends State<PeekAndPopController> with TickerProviderStateMixin {
  final Widget uiChild;
  final PeekAndPopBuilder peekAndPopBuilder;
  final bool useCache;
  final Color backdropColor;
  final Widget overlayBuilder;
  
  final bool isHero;
  
  final PeekAndPopCallback onPeekAndPopComplete;
  final PeekAndPopCallback onPushPeekAndPop;
  final PeekAndPopCallback onUpdatePeekAndPop;
  final PeekAndPopCallback onCancelPeekAndPop;
  final PeekAndPopCallback onFinishPeekAndPop;
  final PeekAndPopCallback onClosePeekAndPop;

  final double treshold;
  final double startPressure;
  final double peakPressure;
  final double peekScale;
  final double peekCoefficient;
  final PeekAndPopMoveController moveController;
  
  final Function popTransition;
  
  ///A required precaution for preventing consecutive Peek & Pop processes without sufficient time. I actually forgot why it is required but it is 
  ///required.
  DateTime lastActionTime;

  ///The primary [AnimationController] used to set the values of the [Backdrop] widget and the view during the Peek stage.
  AnimationController animationController;
  ///The secondary [AnimationController] used to create an iOS-Style jump effect during the Pop stage.
  AnimationController secondaryAnimationController;
  ///The [Animation] controlled by the [secondaryAnimationController].
  Animation<double> secondaryAnimation;
  ///[ValueNotifier] for controlling the state of the [PeekAndPopChildState] as the [animationController] or the [secondaryAnimationController] changes.
  ValueNotifier<double> animationTrackerNotifier = ValueNotifier<double>(0.0);

  PeekAndPopChildState peekAndPopChild;

  ///A value for tracking the stage of the Peek & Pop process.
  bool isComplete = false;
  ///A value for tracking the stage of the Peek & Pop process.
  bool isPushed = false;
  //TODO: Use an Enum instead.
  
  ///A required precaution for behaving accordingly if the [uiChild] is tapped instead of pressed.
  bool isDirect = false;
  ///A required precaution for something that I also actually forgot. 
  bool ignoreAnimation = false;
  
  ///A [ValueNotifier] for rerouting gesture recognition from [PeekAndPopDetector] to [PeekAndPopChildState].
  ValueNotifier<bool> pressReroutedNotifier = ValueNotifier<bool>(false);

  ///Set this value to false if you want Long Press to be used instead of Force Press. Long Press version of this package is still under development
  ///and is not fully tested yet so consider it as a developers preview.
  bool supportsForcePress = true;

  ///The callback for resetting the state of the [PeekAndPopDetector] once the gesture recognition is rerouted to the instantiated 
  ///[PeekAndPopChildState] or vice-versa.
  Function callback;
  
  ///A value to determine the depth of debug printing that is actually only here for myself and the Swiss scientists. 
  int _debugLevel = 0;

  PeekAndPopControllerState(
    this.uiChild,
    this.peekAndPopBuilder,
    this.useCache,
    this.backdropColor,
    this.overlayBuilder,
    this.isHero,
    this.onPeekAndPopComplete,
    this.onPushPeekAndPop,
    this.onUpdatePeekAndPop,
    this.onCancelPeekAndPop,
    this.onFinishPeekAndPop,
    this.onClosePeekAndPop,
    this.treshold,
    this.startPressure,
    this.peakPressure,
    this.peekScale,
    this.peekCoefficient,
    this.moveController,
    this.popTransition);

  ///Called by both [animationController] and [secondaryAnimationController] to change the value of the [animationTrackerNotifier].
  void UpdateAnimationTrackerNotifier() {
    animationTrackerNotifier.value += 1;
  }

  ///[AnimationStatus] listener for the [animationController]. It is used to determine the stage of the Peek & Pop process.
  void PrimaryAnimationStatusListener(AnimationStatus animationStatus) {
    if (ignoreAnimation) return;

    switch (animationStatus) {
      case AnimationStatus.completed:
        if (_debugLevel > 1) print("AnimationStatus.completed");

        HapticFeedback.heavyImpact();

        lastActionTime = DateTime.now();
        isComplete = true;
        pressReroutedNotifier.value = false;

        callback();
        break;
      case AnimationStatus.dismissed:
        if (_debugLevel > 1) print("AnimationStatus.dismissed");

        HapticFeedback.heavyImpact();

        lastActionTime = DateTime.now();
        Navigator.of(context).pop();

        callback();
        break;
      default:
        break;
    }
  }

  ///[AnimationStatus] listener for the [secondaryAnimationController]. It is used function to determine reverse the [secondaryAnimationController]
  /// after it plays once for the iOS-Style jump effect.
  void SecondaryAnimationStatusListener(AnimationStatus animationStatus) {
    if (ignoreAnimation) return;

    switch (animationStatus) {
      case AnimationStatus.completed:
        secondaryAnimation = 
            Tween(
              begin: 1.0 - peekScale - peekCoefficient, 
              end: 1.0 - peekScale - peekCoefficient + 0.1)
            .animate(CurvedAnimation(
                      parent: secondaryAnimationController, 
                      curve: Curves.elasticInOut));
        secondaryAnimationController.reverse();
        break;
      default:
        break;
    }
  }
  //TODO: There must be a better way of doing this!
  
  ///This function is called by the instantiated [PeekAndPopChildState] once it is ready to be included in the Peek & Pop process. Perhaps the most
  ///essential functionality of this package also takes places in this function: The gesture recognition is rerouted from the  [PeekAndPopDetector]
  ///to the instantiated [PeekAndPopChildState]. This is important for avoiding the necessity of having the user stop and restart their Force Press.
  ///Instead, the [PeekAndPopControllerState] does this automatically so that the existing Force Press can continue to update even when if
  ///[PeekAndPopDetector] is blocked by the view which is often the case especially when using PlatformViews. 
  void PushComplete(PeekAndPopChildState peekAndPopChild) {
    if (_debugLevel > 1) print("PushComplete");

    this.peekAndPopChild = peekAndPopChild;
    isPushed = true;

    if (!isDirect) {
      pressReroutedNotifier.value = true;
      Future.delayed(Duration(milliseconds: 333), () {
        GestureBinding.instance.StartIgnoring();
        PointerUpEvent pointerUpEvent = PointerUpEvent(
                                          timeStamp: Duration(milliseconds: GestureBinding.instance.lastEvent.timeStamp.inMilliseconds + 100),
                                          pointer: GestureBinding.instance.lastEvent.pointer,
                                          device: GestureBinding.instance.lastEvent.device,
                                          position: GestureBinding.instance.lastEvent.position,
                                          pressure: GestureBinding.instance.lastEvent.pressure,
                                          pressureMax: GestureBinding.instance.lastEvent.pressureMax,
                                          pressureMin: GestureBinding.instance.lastEvent.pressureMin,
                                          distance: GestureBinding.instance.lastEvent.distance,
                                          distanceMax: GestureBinding.instance.lastEvent.distanceMax,
                                          size: GestureBinding.instance.lastEvent.size,
                                          radiusMajor: GestureBinding.instance.lastEvent.radiusMajor,
                                          radiusMinor: GestureBinding.instance.lastEvent.radiusMinor,
                                          radiusMin: GestureBinding.instance.lastEvent.radiusMin,
                                          radiusMax: GestureBinding.instance.lastEvent.radiusMax,
                                          orientation: GestureBinding.instance.lastEvent.orientation,
                                          tilt: GestureBinding.instance.lastEvent.tilt,
                                          transform: GestureBinding.instance.lastEvent.transform);
        GestureBinding.instance.AddToPendingPointerEvents(pointerUpEvent);
        PointerAddedEvent pointerAddedEvent = PointerAddedEvent(
                                                timeStamp: Duration(milliseconds: GestureBinding.instance.lastEvent.timeStamp.inMilliseconds + 100),
                                                device: GestureBinding.instance.lastEvent.device,
                                                position: GestureBinding.instance.lastEvent.position,
                                                pressureMax: GestureBinding.instance.lastEvent.pressureMax,
                                                pressureMin: GestureBinding.instance.lastEvent.pressureMin,
                                                distance: GestureBinding.instance.lastEvent.distance,
                                                distanceMax: GestureBinding.instance.lastEvent.distanceMax,
                                                radiusMin: GestureBinding.instance.lastEvent.radiusMin,
                                                radiusMax: GestureBinding.instance.lastEvent.radiusMax,
                                                orientation: GestureBinding.instance.lastEvent.orientation,
                                                tilt: GestureBinding.instance.lastEvent.tilt,
                                                transform: GestureBinding.instance.lastEvent.transform);
        GestureBinding.instance.AddToPendingPointerEvents(pointerAddedEvent);
        PointerDownEvent pointerDownEvent = PointerDownEvent(
                                              timeStamp: Duration(milliseconds: GestureBinding.instance.lastEvent.timeStamp.inMilliseconds + 100),
                                              pointer: GestureBinding.instance.lastEvent.pointer,
                                              device: GestureBinding.instance.lastEvent.device,
                                              position: GestureBinding.instance.lastEvent.position,
                                              pressure: GestureBinding.instance.lastEvent.pressure,
                                              pressureMax: GestureBinding.instance.lastEvent.pressureMax,
                                              pressureMin: GestureBinding.instance.lastEvent.pressureMin,
                                              distanceMax: GestureBinding.instance.lastEvent.distanceMax,
                                              size: GestureBinding.instance.lastEvent.size,
                                              radiusMajor: GestureBinding.instance.lastEvent.radiusMajor,
                                              radiusMinor: GestureBinding.instance.lastEvent.radiusMinor,
                                              radiusMin: GestureBinding.instance.lastEvent.radiusMin,
                                              radiusMax: GestureBinding.instance.lastEvent.radiusMax,
                                              orientation: GestureBinding.instance.lastEvent.orientation,
                                              tilt: GestureBinding.instance.lastEvent.tilt,
                                              transform: GestureBinding.instance.lastEvent.transform);
        GestureBinding.instance.AddToPendingPointerEvents(pointerDownEvent);
        GestureBinding.instance.StopIgnoring();
      });
    }

    if (onPushPeekAndPop != null) onPushPeekAndPop(this);
  }

  @override
  void initState() {
    super.initState();

    lastActionTime = DateTime.now();

    animationController = 
      AnimationController(
        vsync: this, 
        duration: const Duration(milliseconds: 333), 
        lowerBound: 0, 
        upperBound: 1)
      ..addListener(UpdateAnimationTrackerNotifier)
      ..addStatusListener(PrimaryAnimationStatusListener);
    secondaryAnimationController = 
      AnimationController(
        vsync: this, 
        duration: const Duration(milliseconds: 166), 
        lowerBound: 0, 
        upperBound: 1)
      ..addListener(UpdateAnimationTrackerNotifier)
      ..addStatusListener(SecondaryAnimationStatusListener);
    secondaryAnimation = 
      Tween(
        begin: 0.0, 
        end: 1.0 - peekScale - peekCoefficient + 0.1)
      .animate(CurvedAnimation(
                parent: secondaryAnimationController, 
                curve: Curves.elasticInOut));
  }

  @override
  void dispose() {
    animationController.removeListener(UpdateAnimationTrackerNotifier);
    animationController.removeStatusListener(PrimaryAnimationStatusListener);
    animationController.dispose();
    secondaryAnimation.removeListener(UpdateAnimationTrackerNotifier);
    secondaryAnimationController.removeStatusListener(SecondaryAnimationStatusListener);
    secondaryAnimationController.dispose();
    
    super.dispose();
  }

  ///This function is called if the [uiChild] is tapped instead of pressed.
  void OnPeekAndPopComplete() {
    if (isComplete || DateTime.now().difference(lastActionTime).inSeconds < 1) return;

    if (_debugLevel > 0) print("OnPeekAndPopComplete");

    isComplete = true;
    isDirect = true;
    ignoreAnimation = true;
    animationController.value = 1;

    Navigator.of(context).push(PeekAndPopRoute(this,(BuildContext context) => PeekAndPopChild(this), popTransition))
    .whenComplete(() {
      HapticFeedback.heavyImpact();

      lastActionTime = DateTime.now();
      Future.delayed(Duration(milliseconds: 666), () {
        animationController.value = 0;
        secondaryAnimationController.value = 0;
        secondaryAnimation = 
          Tween(
              begin: 0.0, 
              end: 1.0 - peekScale - peekCoefficient + 0.1)
          .animate(CurvedAnimation(
                    parent: secondaryAnimationController, 
                    curve: Curves.elasticInOut));
        animationTrackerNotifier.value = 0;
        peekAndPopChild = null;
        isComplete = false;
        isPushed = false;
        isDirect = false;
        ignoreAnimation = false;
        moveController?.Reset();
        pressReroutedNotifier.value = false;
      });

      if (onPeekAndPopComplete != null) onPeekAndPopComplete(this);
    });
  }

  ///Initial stage of the Peek & Pop process.
  void PushPeekAndPop(dynamic pressDetails) {
    if (isComplete || DateTime.now().difference(lastActionTime).inSeconds < 1) return;

    if (_debugLevel > 0) print("PushPeekAndPop");

    Navigator.of(context).push(PeekAndPopRoute(this, (BuildContext context) => PeekAndPopChild(this), popTransition))
    .whenComplete(() {
      HapticFeedback.heavyImpact();

      lastActionTime = DateTime.now();
      Future.delayed(Duration(milliseconds: 666), () {
        animationController.value = 0;
        secondaryAnimationController.value = 0;
        secondaryAnimation = 
          Tween(
            begin: 0.0, 
            end: 1.0 - peekScale - peekCoefficient + 0.1)
            .animate(CurvedAnimation(
                      parent: secondaryAnimationController, 
                      curve: Curves.elasticInOut));
        animationTrackerNotifier.value = 0;
        peekAndPopChild = null;
        isComplete = false;
        isPushed = false;
        isDirect = false;
        ignoreAnimation = false;
        moveController?.Reset();
        pressReroutedNotifier.value = false;
      });
      
      if (onPeekAndPopComplete != null) onPeekAndPopComplete(this);
    });

    if (supportsForcePress) JumpPeekAndPop(pressDetails.pressure);
  }

  ///Main stage of the Peek & Pop process.
  void UpdatePeekAndPop(dynamic pressDetails, {bool isFromOverlayEntry: false}) {
    if (isComplete || DateTime.now().difference(lastActionTime).inSeconds < 1) return;
    if (!supportsForcePress) return;
    if (secondaryAnimationController.isAnimating) return;

    if (_debugLevel > 2) {
      if (!isFromOverlayEntry)
        print("PeekAndPopController: " + pressDetails.pressure.toString());
      else
        print("PeekAndPopChild: " + pressDetails.pressure.toString());
    }

    JumpPeekAndPop(pressDetails.pressure);
    if (pressDetails.pressure > treshold && peekAndPopChild?.animationController?.value == 0) peekAndPopChild.animationController.forward(from: 0.5);
    else if(moveController?.canSetMoveOffset){
      moveController.canSetMoveOffset= false;
      Future.delayed(Duration(milliseconds: 333), () {
        moveController.currentMoveOffset = Offset(-1,-1);
      });
    }
    if(moveController?.currentMoveOffset != null){
      if(moveController.currentMoveOffset.dx == -1 || moveController.currentMoveOffset.dy == -1) {
        moveController.initialMoveOffset = pressDetails.localPosition;
        moveController.currentMoveOffset= pressDetails.localPosition;
      }
      else {
        moveController.currentMoveOffset = pressDetails.localPosition;
        
        if(moveController.callback!=null) moveController.callback(moveController.delta);
      }
    }

    if (onUpdatePeekAndPop != null) onUpdatePeekAndPop(this);
  }

  ///Cancellation stage of the Peek & Pop process.
  void CancelPeekAndPop(dynamic pressDetails, {bool isFromOverlayEntry: false}) {
    if (isComplete || !isPushed || DateTime.now().difference(lastActionTime).inSeconds < 1) return;

    if (_debugLevel > 0) {
      if (!isFromOverlayEntry)
        print("PeekAndPopController: CancelPeekAndPop");
      else
        print("PeekAndPopChild: CancelPeekAndPop");
    }

    DrivePeekAndPop(false);
    if (peekAndPopChild?.animationController?.value != 0) peekAndPopChild.animationController.reverse();

    if (onCancelPeekAndPop != null) onCancelPeekAndPop(this);
  }

  ///Conclusion stage of the Peek & Pop process.
  void FinishPeekAndPop(dynamic pressDetails, {bool isFromOverlayEntry: false}) {
    if (isComplete || DateTime.now().difference(lastActionTime).inSeconds < 1) return;
    if(moveController != null && moveController.ignorePeakWhenMoved && moveController.canIgnorePeak) return;
    
    DrivePeekAndPop(true);

    if (onFinishPeekAndPop != null) onFinishPeekAndPop(this);
  }
  
  ///To close the Peek & Pop process, call this function instead of using the Navigator yourself.
  void ClosePeekAndPop() {
    if (!isComplete || !isPushed || DateTime.now().difference(lastActionTime).inSeconds < 1) return;

    if (_debugLevel > 0) print("ClosePeekAndPop");

    ignoreAnimation = true;
    Navigator.of(context).pop();

    if (onClosePeekAndPop != null) onClosePeekAndPop(this);
  }
  
  void DrivePeekAndPop(bool forward) {
    if (_debugLevel > 0) print("DrivePeekAndPop: $forward");

    if (forward) {
      animationController.forward(from: animationController.value);
      secondaryAnimationController.forward();
      if(!useCache) peekAndPopChild.setState(() {});
    } 
    else {
      animationController.reverse(from: animationController.value);
      secondaryAnimationController.reverse();
    }
  }

  void JumpPeekAndPop(double value) {
    if (_debugLevel > 3) print("JumpPeekAndPop: $value");

    if(value == 1 ) FinishPeekAndPop(null);
    else animationController.value = value;
  }

  @override
  Widget build(BuildContext context) {
    return PeekAndPopDetector(this, uiChild);
  }
}
